// parser1.cpp : Main source file for the parser implementation
//

#ifdef _WIN32
#include "stdafx.h"
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <BipCpi.h>
#include <BipCni.h>
#include "ParserUtils.h"
#include "NodeUtils.h"

#define	SEGMENT_TERMINATOR	0x0D


#ifdef _WIN32
#define	LOG_FILE "C:\\HL7.txt"
#else
#define LOG_FILE "/tmp/HL7.txt"
#endif

///////////////////////////////////////////////////////////
// Context
struct Context
{
	int	iOffset;
	int iClaimedSize;
};
typedef struct Context Context;

#ifdef _WIN32
void __declspec(dllexport) *bipGetParserFactory();



BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}
#endif

FILE	*logFile = 0;

///////////////////////////////////////////////////////////
// ErrorToString
char *ErrorToString(int iRc)
{
	static char	message[256];
	switch(iRc)
	{
	case CCI_PARSER_NAME_TOO_LONG:
		return("Parse name too long");
	default:
		sprintf(message, "Error: Unknown code: %d", iRc);
		return(message);
	}
}
///////////////////////////////////////////////////////////
// Log
void Log(char *szFormat, ...)
{
#ifndef _DEBUG
	return;
#else

	va_list	list;
	va_start(list, szFormat);
	if (logFile == NULL)
	{
		logFile = fopen(LOG_FILE, "a");
		if (logFile == NULL)
		{
			return;
		}
	}
	vfprintf(logFile, szFormat, list);
	fflush(logFile);
	va_end(list);
#endif
}

///////////////////////////////////////////////////////////
// AddNameElement
CciElement *AddNameElement(
				CciParser	*parser,
				CciElement	*pRoot,
				char		*pNameA)
{
	CciElement	*pElement;
	int			iRc;
	CciChar		*pNameU;

	pNameU = mqsixASCIIToUnicode(pNameA);
	pElement = cpiCreateAndInitializeElement(&iRc,
		parser,
		CCI_ELEMENT_TYPE_NAME,
		pNameU,
		1,1);
	free(pNameU);
	cpiAddAsLastChild(&iRc, pRoot, pElement);
	return(pElement);
}

///////////////////////////////////////////////////////////
// AddNameValueElement
CciElement *AddNameValueElement(CciParser	*parser,
								CciElement	*pRoot,
								char		*pNameA,
								char		*pValueA)
{
	CciElement	*pElement;
	int			iRc;
	CciChar		*pStringU;
	CciChar		*pNameU;
	int			iLength;

	pNameU = mqsixASCIIToUnicode(pNameA);
	pElement = cpiCreateAndInitializeElement(&iRc,
		parser,
		CCI_ELEMENT_TYPE_NAME_VALUE,
		pNameU,
		1,1);
	free(pNameU);
	pStringU = mqsixASCIIToUnicode(pValueA);
	if (pStringU != NULL)
	{
		iLength = strlen(pValueA);
		if (iLength == 0)
		{
			iLength = 1;
		}
		cpiSetElementCharacterValue(&iRc,
			pElement,
			pStringU,
			iLength);
		if (iRc != 0)
		{
			Log("Error: iRc=%d, %s[%d]\n", iRc, __FILE__, __LINE__);
		}
		free(pStringU);
	}
	else
	{
		Log("Error: %s[%d]\n", __FILE__, __LINE__);
	}
	cpiAddAsLastChild(&iRc, pRoot, pElement);
	return(pElement);
}

///////////////////////////////////////////////////////////
// CreateContext
void *CreateContext(CciParser* parser)
{
	Context	*pContext;
	Log(">> CreateContext\n");
	pContext = (Context *) malloc(sizeof(Context));
	Log("<< CreateContext\n");
	return(pContext);
}

///////////////////////////////////////////////////////////
// Delete Context
void DeleteContext(CciParser*  parser, CciContext* context)
{
	Log(">> Entered DeleteContext\n");
	if (context != NULL)
	{
		free(context);
	}
	Log("<< Entered DeleteContext\n");
}


///////////////////////////////////////////////////////////
// ParseBuffer
int ParseBuffer(CciParser	*parser,
			    CciContext	*context,
			    int         iOffset)
{
	CciSize iBufferSize;
	Context	*pContext;
	int		iClaimedSize;
	int		iRc;



	Log(">> Entering ParseBuffer()\n");

	pContext = (Context *)context;

	iBufferSize = cpiBufferSize(&iRc, parser);
	iClaimedSize = iBufferSize - iOffset;
	Log("-- offset=%d, buffser size=%d\n", iOffset, iBufferSize);
	Log("-- claiming %d bytes\n", iClaimedSize);
	pContext->iClaimedSize = iClaimedSize;
	pContext->iOffset;
	Log("<< Leaving ParseBuffer()\n");
	return(iClaimedSize);
}


/*

                                 Root
								  |
							     HL7
								  |
						+---------+-------+------------------+
						|				  |					 |
				<Segment Name>    <Segment Name> ... <Segment Name>
					 |					  |					 |
            +--------+-------+		   +-----+			  +-----+
            |		 |		 |		   |	 |			  |		|
          FIELD    FIELD   FIELD     FIELD FIELD        FIELD FIELD

If field contains no components, then FIELD is

	NAME/VALUE where 
	NAME = "Field"
	VALUE = <Value of field>

If field contains components, then FIELD is

				FIELD
				  |
			+-----+-----+
			|			|
		COMPONENT   COMPONENT ...

Where
FIELD NAME = "Field"
and COMPONENT = "NAME/VALUE" where
NAME = "Component"
VALUE = <Value of component>


Special case:
When SEGMENT name is MSH,
it contains a child called "Delimiters" which has the following children:

Field
Component
Repitition
Escape
Subcomponent

*/

void OutputField(char *strField)
{
	Log("Field: \"%s\"\n", strField);
}
void OutputComponent(char *strComponent)
{
	Log("\tComponent: \"%s\"\n", strComponent);
}
void OutputSegmentName(char *strSegmentName)
{
	Log("Segment Name: \"%s\"\n", strSegmentName);
}

///////////////////////////////////////////////////////////
// CreateSegment
CciElement *CreateSegment(CciParser *pParser, CciElement *pRoot, char *name)
{
	OutputSegmentName(name);
	return(AddNameElement(pParser, pRoot, name));
}

///////////////////////////////////////////////////////////
// AddDelimiters
// Add the following NAME/VALUE nodes subordinate to root with values
// of the delimiter/separator characters supplied as parameters:
//
// o Field
// o Component
// o Repitition
// o Escape
// o Subcomponent
//
// The idea here will be to create the following tree:
//
//								Root
//								  |
//								 HL7
//								  |
//								 MSH
//								  |
//							  Delimiters
//								  |
//			+-------+-----------+-+--------+--------+
//			|		|			|		   |		|
//		  Field	Component	Repitition	Escape	Subcomponent
//
CciElement *AddDelimeters(CciParser *pParser, CciElement *pRoot,
	char	fieldSeparator,
	char	componentSeparator,
	char	subcomponentSeparator,
	char	repititionSeparator,
	char	escapeSeparator)
{
	CciElement	*pDelimitersElement;
	char		buffer[2];

	buffer[1] = '\0';
	pDelimitersElement = AddNameElement(pParser, pRoot, "Delimiters");
	buffer[0] = fieldSeparator;
	AddNameValueElement(pParser, pDelimitersElement, "Field", buffer);
	buffer[0] = componentSeparator;
	AddNameValueElement(pParser, pDelimitersElement, "Component", buffer);
	buffer[0] = repititionSeparator;
	AddNameValueElement(pParser, pDelimitersElement, "Repitition", buffer);
	buffer[0] = escapeSeparator;
	AddNameValueElement(pParser, pDelimitersElement, "Escape", buffer);
	buffer[0] = subcomponentSeparator;
	AddNameValueElement(pParser, pDelimitersElement, "Subcomponent", buffer);
	return(pDelimitersElement);
}


///////////////////////////////////////////////////////////
// Process a field of HL7 Data ...
void ProcessField(CciParser		*pParser,
				  CciElement	*pRoot,
				  char			*pField,
				  char			componentSeparator)
{
	char		*pTemp = pField;
	char		*pStart = pField;
	int			iHasComponents = 0;		// 0 - Field has NO components
	char		buffer[512];
	int			iLength;
	CciElement	*pElement;

	// Ensure that a good field was supplied ...
	if (pField == NULL)
	{
		Log("NULL field passed!!! %s[%d]\n", __FILE__, __LINE__);
		return;
	}

	Log("\t>> ProcessField: \"%s\" sep='%c'\n", pField, componentSeparator);

	// Loop through all the characters in the field
	while(*pTemp != '\0')
	{
	// Have we found a component separator?
		if (*pTemp == componentSeparator)
		{
	// Yes!! Is this the first component in the field that we have found?  If
	// so, add a NAME node called "Field" that will serve as the parent of all
	// subsequent components in this field.
			if (iHasComponents == 0)
			{
				pElement = AddNameElement(pParser, pRoot, "Field");
				iHasComponents = 1;
			}
			iLength = pTemp-pStart;				// Length of the component in chars
			memcpy(buffer, pStart, iLength);	// Build a string that is just the component
			buffer[iLength] = '\0';				// NULL terminate the string
			OutputComponent(buffer);			// Debugging
			AddNameValueElement(pParser, pElement, "Component", buffer);
			pStart = pTemp + 1;
		}
		pTemp++;
	}

	// We have processed all the characters in the field ...
	// If we previously found a component, add the last component because
	// ther MUST be one.
	if (iHasComponents)
	{
		iLength = pTemp-pStart;
		memcpy(buffer, pStart, iLength);
		buffer[iLength] = '\0';
		OutputComponent(buffer);
		AddNameValueElement(pParser, pElement, "Component", buffer);
	}
	else	// No components in the field ... output whole field
	{
		OutputField(pField);
		AddNameValueElement(pParser, pRoot, "Field", pField);
	}
	Log("\t<<ProcessField\n");
}

///////////////////////////////////////////////////////////
// Parse a complete HL7 message
void ParseData(char *data, int iLength,
			   CciParser *pParser, CciElement *pRoot)
{
	char	fieldSeparator;
	char	componentSeparator;
	char	subcomponentSeparator;
	char	repititionSeparator;
	char	escapeCharacter;

	char	buffer[256];
	char	iCount;

	CciElement	*pCurrentSegment;


	enum state {NONE, STARTING_MSH, FIELD} state;

	state = NONE;
	iCount = 0;

	while(iLength > 0)
	{
		if (*data == SEGMENT_TERMINATOR)
		{
			switch(state)
			{
			case NONE:
				// Do nothing ...
				break;
			case FIELD:
				buffer[iCount] = '\0';
				ProcessField(pParser, pCurrentSegment, buffer, componentSeparator);
				iCount = 0;
				state = NONE;
				goto end;
				break;
			}
		}
		if (state == NONE)
		{
			if (iCount < 3)
			{
				buffer[iCount] = *data;
				iCount++;
			}
			else	// Got 3 characters worth
			{
				if (memcmp(buffer, "MSH", 3) == 0)
				{
					state = STARTING_MSH;
				}
				else
				{
					state = FIELD;
				}
				buffer[iCount] = '\0';
				iCount = 0;
				pCurrentSegment = CreateSegment(pParser, pRoot, buffer);
				if (state == FIELD)
				{
					goto end;
				}
			}
		} // End STATE == NONE

		if (state == STARTING_MSH)
		{
			if (iCount < 5)
			{
				switch(iCount)
				{
				case 0:
					fieldSeparator = *data;
					Log("Field Separator: '%c'\n", fieldSeparator);
					break;
				case 1:
					componentSeparator = *data;
					Log("Component Separator: '%c'\n", componentSeparator);
					break;
				case 2:
					repititionSeparator = *data;
					Log("Repition Separator: '%c'\n", repititionSeparator);
					break;
				case 3:
					escapeCharacter = *data;
					Log("Escape Separator: '%c'\n", escapeCharacter);
					break;
				case 4:
					subcomponentSeparator = *data;
					Log("Subcomponent Separator: '%c'\n", subcomponentSeparator);
					break;
				}
				iCount++;
			}
			else // Skip to the first field separator ...
			{
				if (*data == fieldSeparator)
				{
					AddDelimeters(pParser, pCurrentSegment,
						fieldSeparator,
						componentSeparator,
						subcomponentSeparator,
						repititionSeparator,
						escapeCharacter);
					state = FIELD;
					iCount = 0;
					goto end;
				}
			}
		} // End STATE == STARTING_MSH

		if (state == FIELD)
		{
			if (*data == fieldSeparator)
			{
				buffer[iCount] = '\0';
				ProcessField(pParser, pCurrentSegment, buffer, componentSeparator);
				iCount = 0;
				goto end;
			}
			buffer[iCount] = *data;
			iCount++;
		}

end:
		iLength--;
		data++;
	}

}

///////////////////////////////////////////////////////////
// ParseFirstChild
void ParseFirstChild(CciParser	*parser,
					 CciContext	*context,
					 CciElement	*currentElement)
{
	int		iRc;
	const CciByte	*pPtr;
	CciSize			iSize;
	CciElement		*pRoot;

	Log(">> ParseFirstChild() element=%p\n", currentElement);
	pPtr = cpiBufferPointer(&iRc, parser);
	pRoot = cpiRootElement(&iRc, parser);
	iSize = cpiBufferSize(&iRc, parser);
	ParseData((char *)pPtr, iSize, parser, pRoot);
	cpiSetElementCompleteNext(&iRc, pRoot, 1);
	cpiSetElementCompletePrevious(&iRc, pRoot, 1);


	Log("<< ParseFirstChild()\n");
}

void ParseLastChild(CciParser	*parser,
					   CciContext	*context,
					   CciElement	*currentElement)
{
	Log(">> ParseLastChild()\n");
}

void ParsePreviousSibling(CciParser		*parser,
						  CciContext	*context,
						  CciElement	*currentElement)
{
	Log(">> ParsePreviousSibling()\n");
}


void ParseNextSibling(CciParser	*parser,
						CciContext	*context,
						CciElement	*currentElement)
{
	Log(">> ParseNextSibling()\n");
}


///////////////////////////////////////////////////////////
// WriteBuffer
//
int WriteBuffer(CciParser	*pParser,
				CciContext	*context)
{
	CciElement		*pSegmentElement;
	CciElement		*pDelimitersElement;
	CciElement		*pElement;
	CciElement		*pFieldElement;
	CciElement		*pComponentElement;
	CciElementType	iType;

	int			iRc;
	char		*strSegmentNameA;
	char		fieldSeparator;
	char		componentSeparator;
	char		subcomponentSeparator;
	char		repititionSeparator;
	char		escapeSeparator;
	char		*pValueA;

	Log(">> WriteBuffer()\n");
	// Process each child element ...

	pSegmentElement = cpiFirstChild(&iRc, cpiRootElement(&iRc, pParser));

	while(pSegmentElement != NULL)
	{
		strSegmentNameA = GetName(pParser, pSegmentElement);
		Log("Found segment \"%s\"\n", strSegmentNameA);
	// Output the 3 character segment name
		cpiAppendToBuffer(&iRc, pParser, strSegmentNameA, 3);
		if (strcmp(strSegmentNameA, "MSH") == 0)
		{
			pDelimitersElement = cniSearchFirstChild(&iRc,
				pSegmentElement,
				CCI_COMPARE_MODE_NAME,
				0,
				L"Delimiters");
			if (pDelimitersElement != NULL)
			{
				// Get the Field Separator
				pElement = cniSearchFirstChild(&iRc,
					pDelimitersElement,
					CCI_COMPARE_MODE_NAME,
					0,
					L"Field");
				if (pElement != 0)
				{
					pValueA = GetStringValue(pParser, pElement);
					fieldSeparator = pValueA[0];
					Log("Field delimiter: '%c'\n", fieldSeparator);
					free(pValueA);
				}

				// Get the Component Separator
				pElement = cniSearchFirstChild(&iRc,
					pDelimitersElement,
					CCI_COMPARE_MODE_NAME,
					0,
					L"Component");
				if (pElement != 0)
				{
					pValueA = GetStringValue(pParser, pElement);
					componentSeparator = pValueA[0];
					Log("Component delimiter: '%c'\n", componentSeparator);
					free(pValueA);
				}
				// Get the Repitition Separator
				pElement = cniSearchFirstChild(&iRc,
					pDelimitersElement,
					CCI_COMPARE_MODE_NAME,
					0,
					L"Repitition");
				if (pElement != 0)
				{
					pValueA = GetStringValue(pParser, pElement);
					repititionSeparator = pValueA[0];
					free(pValueA);
					Log("Repitition delimiter: '%c'\n", repititionSeparator);
				}

				// Get the Escape Separator
				pElement = cniSearchFirstChild(&iRc,
					pDelimitersElement,
					CCI_COMPARE_MODE_NAME,
					0,
					L"Escape");
				if (pElement != 0)
				{
					pValueA = GetStringValue(pParser, pElement);
					escapeSeparator = pValueA[0];
					free(pValueA);
					Log("Escape delimiter: '%c'\n", escapeSeparator);
				}

				// Get the Subcomponent Separator
				pElement = cniSearchFirstChild(&iRc,
					pDelimitersElement,
					CCI_COMPARE_MODE_NAME,
					0,
					L"Subcomponent");
				if (pElement != 0)
				{
					pValueA = GetStringValue(pParser, pElement);
					subcomponentSeparator = pValueA[0];
					free(pValueA);
					Log("Subcomponent delimiter: '%c'\n", subcomponentSeparator);
				}
			}
			cpiAppendToBuffer(&iRc, pParser, &fieldSeparator, 1);
			cpiAppendToBuffer(&iRc, pParser, &componentSeparator, 1);
			cpiAppendToBuffer(&iRc, pParser, &subcomponentSeparator, 1);
			cpiAppendToBuffer(&iRc, pParser, &repititionSeparator, 1);
			cpiAppendToBuffer(&iRc, pParser, &escapeSeparator, 1);
		} // End of "Is Segment MSH?"
		free(strSegmentNameA);
		cpiAppendToBuffer(&iRc, pParser, &fieldSeparator, 1);
		// Now ... write out all the fields ...
		pFieldElement = cpiFirstChild(&iRc, pSegmentElement);
		while(pFieldElement != NULL)
		{
			iType = cpiElementType(&iRc, pFieldElement);
	// If the node type is NAME only, then it contains subordinates which
	// are components.
			if (iType == CCI_ELEMENT_TYPE_NAME)
			{
				pComponentElement = cpiFirstChild(&iRc, pFieldElement);
				while(pComponentElement != NULL)
				{
					pValueA = GetStringValue(pParser, pFieldElement);
					if (pValueA != NULL)
					{
						Log("Appending subcomponent \"%s\"\n", pValueA);
						cpiAppendToBuffer(&iRc, pParser, pValueA, strlen(pValueA));
						free(pValueA);
					}
					else
					{
						Log("Component value of NULL!\n");
					}
					pComponentElement = cpiNextSibling(&iRc, pComponentElement);
					if (pComponentElement != NULL)
					{
						cpiAppendToBuffer(&iRc, pParser, &componentSeparator, 1);
					}
				}
			}
	// ... else ... this will be a field with no components if its type is NAME/VALUE
			else if (iType == CCI_ELEMENT_TYPE_NAME_VALUE)
			{
				pValueA = GetStringValue(pParser, pFieldElement);
				if (pValueA != NULL)
				{
					Log("Appending field \"%s\"\n", pValueA);
					cpiAppendToBuffer(&iRc, pParser, pValueA, strlen(pValueA));
					free(pValueA);
				}
				else
				{
					Log("Field value of NULL!\n");
				}
				cpiAppendToBuffer(&iRc, pParser, &fieldSeparator, 1);
			}
			pFieldElement = cpiNextSibling(&iRc, pFieldElement);
		}	// End of all fields ...
		cpiAppendToBuffer(&iRc, pParser, "\015", 1);
		pSegmentElement = cpiNextSibling(&iRc, pSegmentElement);
	}
	Log("<< WriteBuffer()\n");
	return(0);
}


///////////////////////////////////////////////////////////
// bipGetParserFactory
void *bipGetParserFactory()
{
	CciFactory*     factoryObject;
	int             rc;
	static CPI_VFT  vftable;
	
	Log("*** HL7 Build: %s ***\n", __TIMESTAMP__);
	Log("%s", "Entered bipGetParserFactory()\n");
	/* Setup virtual function table with addresses of implementation functions */
	memset(&vftable, 0, sizeof(vftable));
	vftable.iFpCreateContext			= CreateContext;
	vftable.iFpDeleteContext			= DeleteContext;
	vftable.iFpParseBuffer				= ParseBuffer;
	vftable.iFpParseFirstChild			= ParseFirstChild;
	vftable.iFpParseLastChild			= ParseLastChild;
	vftable.iFpParseLeftSibling			= ParsePreviousSibling;
	vftable.iFpParseRightSibling		= ParseNextSibling;
	vftable.iFpWriteBuffer				= WriteBuffer;
	

	/* Create the message factory for this plug-in */
	factoryObject = cpiCreateParserFactory(&rc, L"HL7Factory");
	if (rc != CCI_SUCCESS)
	{
		Log("Failed to create parser factory: rc=%d\n", rc);
	}
	if (factoryObject)
	{

		/* Define the classes of message supported by the factory */
		cpiDefineParserClass(&rc, factoryObject, L"HL7", &vftable);
		if (rc != CCI_SUCCESS)
		{
			Log("Failed to create a parser: rc=%d, %s\n", rc, ErrorToString(rc));
		}
		else
		{
			Log("Created the parser called %s\n", "HL7");
		}
	}
	else
	{
		Log("Failed to create factory object in bipGetParserFactory()\n");
		/* Error recovery */ 
	}

	/* Return address of this factory object to the broker */
	return(factoryObject);
}

